"""Test mixins."""

import unittest

from pydantic import BaseModel
from typing_extensions import Self

import curies
from curies import Converter, Reference
from curies.mixins import (
    SemanticallyProcessable,
    SemanticallyStandardizable,
    process_many,
    standardize_many,
)


class TestMixins(unittest.TestCase):
    """Test mixins."""

    def test_semantically_processable(self) -> None:
        """Test processing."""

        class Processed(BaseModel):
            """A processed model, with a reference."""

            reference: Reference

        class Raw(BaseModel, SemanticallyProcessable[Processed]):
            """A raw model, with a URI."""

            uri: str

            def process(self, converter: Converter) -> Processed:
                """Process the raw model."""
                return Processed(reference=converter.parse(self.uri, strict=True).to_pydantic())

        converter = Converter.from_prefix_map({"GO": "http://purl.obolibrary.org/obo/GO_"})
        initial = Raw(uri="http://purl.obolibrary.org/obo/GO_1234567")
        actual = initial.process(converter)
        self.assertIsInstance(actual, Processed)
        expected = Processed(reference=Reference(prefix="GO", identifier="1234567"))
        self.assertEqual(expected, actual)
        self.assertEqual([expected], process_many([initial], converter))

        # check none handling
        self.assertIsNone(process_many(None, converter))

    def test_standardizable(self) -> None:
        """Test standardizable."""

        class HoldsReference(BaseModel, SemanticallyStandardizable):
            """A test class with a reference."""

            reference: Reference

            def standardize(self, converter: Converter) -> Self:
                """Standardize the reference in the object."""
                return self.model_copy(
                    update={
                        "reference": converter.standardize_reference(self.reference, strict=True),
                    }
                )

        converter = Converter(
            [
                curies.Record(
                    prefix="TEST", prefix_synonyms=["test"], uri_prefix="https://example.org/"
                )
            ]
        )
        initial = HoldsReference(reference=Reference(prefix="test", identifier="1234567"))
        actual = initial.standardize(converter)
        expected = HoldsReference(reference=Reference(prefix="TEST", identifier="1234567"))
        self.assertEqual(expected, actual)
        self.assertEqual([expected], standardize_many([initial], converter))

        # check none handling
        self.assertIsNone(standardize_many(None, converter))



"""Dataframe operations."""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Collection
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, TypeGuard, Union, cast

from .utils import _prefix_from_curie

if TYPE_CHECKING:
    import pandas as pd

    from .api import Converter

__all__ = [
    "PrefixIndexMethod",
    "filter_df_by_curies",
    "filter_df_by_prefixes",
    "get_df_curies_index",
    "get_df_prefixes_index",
    "get_df_unique_prefixes",
    "get_filter_df_by_curies_index",
    "get_filter_df_by_prefixes_index",
]

DataframeOrSeries: TypeAlias = Union["pd.DataFrame", "pd.Series[str]"]


def _get_prefix_checker(prefix: str | Collection[str]) -> Callable[[str], bool]:
    """Get a function that checks if a CURIE starts with a prefix."""
    if isinstance(prefix, str):
        prefix_with_colon = prefix + ":"

        def _func(curie: str) -> bool:
            return curie.startswith(prefix_with_colon)

    else:
        prefixes_with_colons = {p + ":" for p in prefix}

        def _func(curie: str) -> bool:
            return any(
                curie.startswith(prefix_with_colon) for prefix_with_colon in prefixes_with_colons
            )

    return _func


def _get_prefixes_from_curie_column(
    df: DataframeOrSeries,
    *,
    column: int | str | None = None,
    converter: Converter | None = None,
    validate: bool,
) -> pd.Series[str]:
    return _get_series(df, column).map(_get_curie_parser(converter=converter, validate=validate))


def _get_curie_parser(
    *, converter: Converter | None = None, validate: bool = False
) -> Callable[[str], str]:
    # TODO what if it can't parse?
    # TODO handle None?
    # TODO handle invalid CURIEs?

    if not validate:
        return _prefix_from_curie
    elif converter is None:
        raise ValueError("converter is required for validation")
    else:

        def _func(curie: str) -> str:
            reference = converter.parse_curie(curie, strict=True)
            return reference.prefix

    return _func


#: The method for filtering on prefixe
PrefixIndexMethod: TypeAlias = Literal["iterative", "precalculated"]


def get_filter_df_by_prefixes_index(
    df: DataframeOrSeries,
    *,
    column: str | int | None = None,
    prefixes: str | Collection[str],
    method: PrefixIndexMethod | None = None,
    validate: bool = False,
    converter: Converter | None = None,
) -> pd.Series[bool]:
    """Get an index of CURIEs in the given column that start with the prefix(es).

    :param df: A dataframe or series. If a dataframe is given, the ``column`` must not be none.
    :param column: The column to check, if a dataframe was passed. If a series was passed, this can be left as none.
    :param prefixes: The prefix or set of prefixes to identify
    :param method: The indexing method
    :param validate: Should the prefixes be validated against the converter?
    :param converter: A converter for validating CURIEs

    :returns: A pandas boolean series that corresponds to the rows of the dataframe or series provided

    :raises ValueError: If validation is set to true but no converter is passed

    Example usage:

    .. code-block:: python

        import pandas as pd
        from curies.dataframe import get_filter_df_by_prefixes_index

        rows = [
            ("DOID:0080795", "skos:exactMatch", "EFO:0003029", "semapv:ManualMappingCuration"),
            ("DOID:0080795", "skos:exactMatch", "mesh:D015471", "semapv:ManualMappingCuration"),
            ("DOID:0080799", "skos:exactMatch", "EFO:1000527", "semapv:ManualMappingCuration"),
            ("DOID:0080808", "skos:exactMatch", "mesh:D000069295", "semapv:ManualMappingCuration"),
        ]
        df = pd.DataFrame(
            rows, columns=["subject_id", "predicate_id", "object_id", "mapping_justification"]
        )
        idx = get_filter_df_by_prefixes_index(df, column="object_id", prefixes=["EFO"])
        filtered_df = df[idx]
    """
    if method == "iterative" or method is None:
        return _get_series(df, column).map(_get_prefix_checker(prefixes))
    elif method == "precalculated":
        if converter is None:  # pragma: no cover
            raise ValueError("a converter is required for method B")
        prefix_series = _get_prefixes_from_curie_column(
            df, column=column, converter=converter, validate=validate
        )
        if isinstance(prefixes, str):
            return prefix_series == prefixes
        else:
            return prefix_series.isin(prefixes)
    else:  # pragma: no cover
        raise ValueError(f"invalid method given: {method}")


def filter_df_by_prefixes(
    df: pd.DataFrame,
    *,
    column: str | int,
    prefixes: str | Collection[str],
    method: PrefixIndexMethod | None = None,
    validate: bool = False,
    converter: Converter | None = None,
) -> pd.DataFrame:
    """Filter a dataframe based on CURIEs in a given column having a given prefix or set of prefixes.

    :param df: A dataframe
    :param column: The integer index or column name of a column containing CURIEs
    :param prefixes: The prefix (given as a string) or collection of prefixes (given as a
        list, set, etc.) to keep
    :param method: The implementation for getting the prefix index
    :param validate: Should the prefixes be validated against the converter?
    :param converter: A converter for validating CURIEs

    :returns: If not in place, return a new dataframe.

    Example usage:

    .. code-block:: python

        import pandas as pd
        from curies.dataframe import filter_df_by_prefixes

        rows = [
            ("DOID:0080795", "skos:exactMatch", "EFO:0003029", "semapv:ManualMappingCuration"),
            ("DOID:0080795", "skos:exactMatch", "mesh:D015471", "semapv:ManualMappingCuration"),
            ("DOID:0080799", "skos:exactMatch", "EFO:1000527", "semapv:ManualMappingCuration"),
            ("DOID:0080808", "skos:exactMatch", "mesh:D000069295", "semapv:ManualMappingCuration"),
        ]
        df = pd.DataFrame(
            rows, columns=["subject_id", "predicate_id", "object_id", "mapping_justification"]
        )
        filtered_df = filter_df_by_prefixes(df, column="object_id", prefixes=["EFO"])

    This results in the following dataframe:

    ============ =============== =========== ============================
    subject_id   predicate_id    object_id   mapping_justification
    ============ =============== =========== ============================
    DOID:0080795 skos:exactMatch EFO:0003029 semapv:ManualMappingCuration
    DOID:0080799 skos:exactMatch EFO:1000527 semapv:ManualMappingCuration
    ============ =============== =========== ============================

    Internally, this function uses :func:`get_filter_df_by_prefixes_index`.
    """
    idx = get_filter_df_by_prefixes_index(
        df=df,
        column=column,
        prefixes=prefixes,
        method=method,
        converter=converter,
        validate=validate,
    )
    return df[idx]


def get_filter_df_by_curies_index(
    df: DataframeOrSeries,
    *,
    column: str | int | None = None,
    curies: str | Collection[str],
) -> pd.Series[bool]:
    """Get an index of CURIEs in the given column that are the given CURIE(s)."""
    series = _get_series(df, column)
    if isinstance(curies, str):
        return series == curies
    else:
        return series.isin(set(curies))


def get_df_curies_index(
    df: DataframeOrSeries, *, column: str | int | None = None
) -> dict[str, list[int]]:
    """Get a dictionary from CURIEs that appear in the column to the row indexes where they appear."""
    dd: defaultdict[str, list[int]] = defaultdict(list)
    for i, curie in enumerate(_get_series(df, column)):
        dd[curie].append(i)
    return dict(dd)


def filter_df_by_curies(
    df: pd.DataFrame,
    *,
    column: str | int,
    curies: str | Collection[str],
) -> pd.DataFrame:
    """Filter a dataframe based on CURIEs in a given column having a given prefix or set of prefixes.

    :param df: A dataframe
    :param column: The integer index or column name of a column containing CURIEs
    :param curies: The CURIE (given as a string) or collection of CURIEs (given as a
        list, set, etc.) to keep

    :returns: If not in place, return a new dataframe.

    Example usage:

    .. code-block:: python

        import pandas as pd
        from curies.dataframe import filter_df_by_curies

        rows = [
            ("DOID:0080795", "skos:exactMatch", "EFO:0003029", "semapv:ManualMappingCuration"),
            ("DOID:0080795", "skos:exactMatch", "mesh:D015471", "semapv:ManualMappingCuration"),
            ("DOID:0080799", "skos:exactMatch", "EFO:1000527", "semapv:ManualMappingCuration"),
            ("DOID:0080808", "skos:exactMatch", "mesh:D000069295", "semapv:ManualMappingCuration"),
        ]
        df = pd.DataFrame(
            rows, columns=["subject_id", "predicate_id", "object_id", "mapping_justification"]
        )
        filtered_df = filter_df_by_curies(df, column="subject_id", prefixes=["DOID:0080795"])

    This results in the following dataframe:

    ============ =============== ============ ============================
    subject_id   predicate_id    object_id    mapping_justification
    ============ =============== ============ ============================
    DOID:0080795 skos:exactMatch EFO:0003029  semapv:ManualMappingCuration
    DOID:0080795 skos:exactMatch mesh:D015471 semapv:ManualMappingCuration
    ============ =============== ============ ============================
    """
    idx = get_filter_df_by_curies_index(df=df, column=column, curies=curies)
    return df[idx]


def get_df_prefixes_index(
    df: DataframeOrSeries,
    *,
    column: str | int | None = None,
    converter: Converter | None = None,
    validate: bool = False,
) -> dict[str, list[int]]:
    """Get a dictionary from prefixes that appear in the column to the row indexes where they appear."""
    dd: defaultdict[str, list[int]] = defaultdict(list)
    f = _get_curie_parser(converter=converter, validate=validate)
    for i, prefix in enumerate(_get_series(df, column).map(f)):
        dd[prefix].append(i)
    return dict(dd)


def get_df_unique_prefixes(
    df: DataframeOrSeries,
    *,
    column: str | int | None = None,
    validate: bool = False,
    converter: Converter | None = None,
) -> set[str]:
    """Get unique prefixes.

    :param df: A dataframe or series. If a dataframe is given, the ``column`` must not be none.
    :param column: The column to check, if a dataframe was passed. If a series was passed, this can be left as none.
    :param validate: Should the prefixes be validated against the converter?
    :param converter: A converter for validating CURIEs

    :returns: A set of prefixes appearing in CURIEs in the given column

    .. code-block:: python

        import pandas as pd
        from curies.dataframe import get_df_unique_prefixes

        rows = [
            ("DOID:0080795", "skos:exactMatch", "EFO:0003029", "semapv:ManualMappingCuration"),
            ("DOID:0080795", "skos:exactMatch", "mesh:D015471", "semapv:ManualMappingCuration"),
            ("DOID:0080799", "skos:exactMatch", "EFO:1000527", "semapv:ManualMappingCuration"),
            ("DOID:0080808", "skos:exactMatch", "mesh:D000069295", "semapv:ManualMappingCuration"),
        ]
        df = pd.DataFrame(
            rows, columns=["subject_id", "predicate_id", "object_id", "mapping_justification"]
        )
        assert get_df_unique_prefixes(df, column="object_id") == {"EFO", "mesh"}
    """
    series = _get_series(df, column)
    f = _get_curie_parser(converter=converter, validate=validate)
    return set(series.map(f).unique())


def _disallowed_dtype(series: pd.Series[Any] | str) -> TypeGuard[pd.Series[str]]:
    import numpy as np

    if isinstance(series, str):
        return False

    return series.dtype != np.str_ and series.dtype != np.dtype("O")


def _get_series(df_or_series: DataframeOrSeries, column: str | int | None = None) -> pd.Series[str]:
    import pandas as pd

    if isinstance(df_or_series, pd.Series):
        if _disallowed_dtype(df_or_series):
            raise TypeError(
                f"passed series that does not have strings: {df_or_series.dtype=} {type(df_or_series.dtype)=}\n\n{df_or_series}"
            )
        return df_or_series

    if column is None:
        raise ValueError("must pass non-none column when using a dataframe directly")

    series = df_or_series[column]
    if _disallowed_dtype(series):
        raise TypeError(
            f"passed series that does not have strings: {series.dtype=} {type(series.dtype)=}\n\n{series}"
        )
    return cast("pd.Series[str]", series)



"""Utilities for working with strings."""

from __future__ import annotations

__all__ = [
    "NoCURIEDelimiterError",
    "_split",
]


class NoCURIEDelimiterError(ValueError):
    """An error thrown on a string with no delimiter."""

    def __init__(self, curie: str):
        """Initialize the error."""
        self.curie = curie

    def __str__(self) -> str:
        return f"{self.curie} does not appear to be a CURIE - missing a delimiter"


def _split(curie: str, *, sep: str = ":") -> tuple[str, str]:
    """Split a CURIE string using string operations."""
    prefix, delimiter, identifier = curie.partition(sep)
    if not delimiter:
        raise NoCURIEDelimiterError(curie)
    return prefix, identifier


def _prefix_from_curie(curie: str, *, sep: str = ":") -> str:
    """Split a CURIE string using string operations and return the prefix."""
    return _split(curie, sep=sep)[0]



"""Mixin classes."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Generic, TypeVar, overload

from typing_extensions import Self

from .api import Converter

__all__ = [
    "SemanticallyProcessable",
    "SemanticallyStandardizable",
    "process_many",
    "standardize_many",
]

X = TypeVar("X")


class SemanticallyProcessable(ABC, Generic[X]):
    """A class that can be processed with a converter.

    The goal of this class is to standardize objects that come with
    unprocessed URIs that can be processed into references with
    respect to a :class:`curies.Converter`. For example, this is
    useful for :mod:`obographs` and :mod:`jskos`.

    .. code-block:: python

        from pydantic import BaseModel
        from curies import SemanticallyProcessable


        class ProcessedEntity(BaseModel):
            reference: Reference


        class RawEntity(BaseModel, SemanticallyProcessable[ProcessedEntity]):
            uri: str

            def process(self, converter: Converter) -> ProcessedEntity:
                return ProcessedEntity(
                    reference=converter.parse_uri(self.uri, strict=True).to_pydantic()
                )
    """

    @abstractmethod
    def process(self, converter: Converter) -> X:
        """Process this raw instance."""
        raise NotImplementedError


# docstr-coverage:excused `overload`
@overload
def process_many(instances: None, converter: Converter) -> None: ...


# docstr-coverage:excused `overload`
@overload
def process_many(
    instances: Iterable[SemanticallyProcessable[X]], converter: Converter
) -> list[X]: ...


def process_many(
    instances: Iterable[SemanticallyProcessable[X]] | None, converter: Converter
) -> list[X] | None:
    """Process multiple semantically processable instances."""
    if instances is None:
        return None
    return [instance.process(converter) for instance in instances]


class SemanticallyStandardizable(ABC):
    """An object that can be standardized.

    In the following example, a simple object is constructed:

    .. code-block:: python

        from typing_extensions import Self
        from curies import Converter, Reference, SemanticallyStandardizable


        class ReferenceHolder(SemanticallyStandardizable):
            def __init__(self, reference):
                self.reference = reference

            def standardize(self, converter: Converter) -> Self:
                return ReferenceHolder(converter.standardize_reference(self.reference, strict=True))

    It's good form to make these operations return new objects, but there's no reason
    you couldn't update the object in place like in :

    .. code-block:: python

        from typing_extensions import Self
        from curies import Converter, Reference, SemanticallyStandardizable


        class ReferenceHolder(SemanticallyStandardizable):
            def __init__(self, reference):
                self.reference = reference

            def standardize(self, converter: Converter) -> Self:
                self.reference = converter.standardize_reference(self.reference, strict=True)
                return self

    In the following example, the :meth:`pydantic.BaseModel.model_copy` is
    used to automatically reuse all other fields that aren't updated, which
    creates a new object.

    .. code-block:: python

        import datetime
        from typing_extensions import Self
        from curies import Converter, Reference, SemanticallyStandardizable
        from pydantic import BaseModel


        class Triple(BaseModel, SemanticallyStandardizable):
            subject: Reference
            predicate: Reference
            object: Reference
            date_asserted: datetime.date

            def standardize(self, converter: Converter) -> Self:
                return self.model_copy(
                    update={
                        "subject": converter.standardize_reference(self.subject, strict=True),
                        "predicate": converter.standardize_reference(self.predicate, strict=True),
                        "object": converter.standardize_reference(self.object, strict=True),
                    }
                )
    """

    @abstractmethod
    def standardize(self, converter: Converter) -> Self:
        """Standardize all references in the object."""
        raise NotImplementedError


Y = TypeVar("Y", bound=SemanticallyStandardizable)


# docstr-coverage:excused `overload`
@overload
def standardize_many(instances: None, converter: Converter) -> None: ...


# docstr-coverage:excused `overload`
@overload
def standardize_many(instances: Iterable[Y], converter: Converter) -> list[Y]: ...


def standardize_many(instances: Iterable[Y] | None, converter: Converter) -> list[Y] | None:
    """Standardize multiple semantically standardizable instances."""
    if instances is None:
        return None
    return [instance.standardize(converter) for instance in instances]



"""External sources of contexts."""

from __future__ import annotations

from typing import Any

from .api import Converter

__all__ = [
    "get_bioregistry_converter",
    "get_go_converter",
    "get_monarch_converter",
    "get_obo_converter",
    "get_prefixcommons_converter",
]

BIOREGISTRY_CONTEXTS = (
    "https://raw.githubusercontent.com/biopragmatics/bioregistry/main/exports/contexts"
)


def get_obo_converter() -> Converter:
    """Get the latest OBO Foundry context.

    :returns: A converter object representing the OBO Foundry's JSON-LD context, which
        contains a simple mapping from OBO Foundry preferred prefixes for ontologies
        that contain case stylization (e.g., ``GO``, not ``go``; ``VariO``, not
        ``vario``).

        It does not include synonyms nor any non-ontology prefixes - e.g., it does not
        include semantic web prefixes like ``rdfs``, it does not include other useful
        biomedical prefixes like ``hgnc``.

    If you want a more comprehensive prefix map, consider using the Bioregistry via
    :func:`get_bioregistry_converter` or by chaining the OBO converter in front of the
    Bioregistry depending on your personal/project preferences using
    :func:`curies.chain`.

    Provenance:

    - This JSON-LD context is generated programmatically by
      https://github.com/OBOFoundry/OBOFoundry.github.io/blob/master/util/processor.py.
    - The file is accessed via from http://purl.obolibrary.org/meta/obo_context.jsonld,
      which is configured through the OBO Foundry's PURL server with
      https://github.com/OBOFoundry/purl.obolibrary.org/blob/master/www/.htaccess and
      ultimately points to
      https://raw.githubusercontent.com/OBOFoundry/OBOFoundry.github.io/master/registry/obo_context.jsonl
    """
    # See configuration on
    # to see where this PURL points
    url = "http://purl.obolibrary.org/meta/obo_context.jsonld"
    return Converter.from_jsonld(url)


def get_prefixcommons_converter(name: str = "monarch_context") -> Converter:
    """Get a Prefix Commons-maintained context.

    :param name: The name of the JSON-LD file (e.g., defaults to ``monarch_context``).
        See the full list at
        https://github.com/prefixcommons/prefixcommons-py/tree/master/prefixcommons/registry.

    :returns: A converter
    """
    url = (
        "https://raw.githubusercontent.com/prefixcommons/prefixcommons-py/master/"
        f"prefixcommons/registry/{name}.jsonld"
    )
    return Converter.from_jsonld(url)


def get_monarch_converter() -> Converter:
    """Get the Prefix Commons-maintained Monarch context."""
    return get_prefixcommons_converter("monarch_context")


def get_go_converter() -> Converter:
    """Get the Prefix Commons-maintained GO context."""
    return get_prefixcommons_converter("go_context")


def get_bioregistry_converter(web: bool = False, **kwargs: Any) -> Converter:
    """Get the latest extended prefix map from the Bioregistry [hoyt2022]_.

    :param web: If false, tries to import :mod:`bioregistry` and use
        :func:`bioregistry.get_converter` to get the converter. Otherwise, falls back to
        using the GitHub-hosted EPM export.
    :param kwargs: Keyword arguments to pass to
        :meth`:curies.Converter.from_extended_prefix_map` when using web-based loading.

    :returns: A converter representing the Bioregistry, which includes a comprehensive
        collection of prefixes, prefix synonyms, and URI prefix synonyms.

    Short summary of the Bioregistry:

    1. It deduplicates and harmonizes dozens of different resources that curate
       partially overlapping and conflicting prefix maps
    2. It contains detailed CURIE prefix synonyms to support standardization
    3. It enforces the generation of a self-consistent extended prefix map

    The Bioregistry's primary prefixes are all standardized to be lowercase, have
    minimal punctuation, and be the most idiomatic possible. When this conflicts with
    your personal preferences/community preferences, you can chain another converter in
    front of the Bioregistry converter using :func:`curies.chain`.

    However, the Bioregistry itself presents a more sustainable way of documenting these
    deviations in a community-oriented way using its "context" configurations. See
    https://bioregistry.io/context/ for more information. One excellent example of a
    community context is for the OBO community (see https://bioregistry.io/context/obo),
    which prioritizes OBO capitalized prefixes and makes a few minor changes for
    backwards compatibility (e.g., renaming Orphanet).

    .. [hoyt2022] `Unifying the identification of biomedical entities with the
        Bioregistry <https://www.nature.com/articles/s41597-022-01807-3>`_
    """
    if not web:
        try:
            import bioregistry
        except ImportError:  # pragma: no cover
            pass
        else:
            return bioregistry.manager.get_converter()
    url = f"{BIOREGISTRY_CONTEXTS}/bioregistry.epm.json"
    return Converter.from_extended_prefix_map(url, **kwargs)



"""Database adapters for :mod:`curies`.

Using :mod:`sqlmodel`
=====================

SQLModel is a joint abstraction over :mod:`pydantic` and :mod:`sqlalchemy`. If you want
to use :class:`curies.Reference` within a SQLModel class, you can do so by setting the
SQLAlchemy column type through the ``sa_column`` keyword argument to the results
returned by :func:`get_reference_sa_column`,

For example, this can be used to model a semantic triple, which has a subject reference,
predicate reference, and object reference like in the following:

.. code-block:: python

    from curies import Reference
    from curies.database import get_reference_sa_column
    from sqlmodel import Field, Session, SQLModel, create_engine, select


    class Edge(SQLModel, table=True):
        id: int | None = Field(default=None, primary_key=True)
        subject: Reference = Field(sa_column=get_reference_sa_column())
        predicate: Reference = Field(sa_column=get_reference_sa_column())
        object: Reference = Field(sa_column=get_reference_sa_column())


    e1 = Edge(subject="CHEBI:135122", predicate="skos:exactMatch", object="mesh:C073738")
    e2 = Edge(subject="CHEBI:135125", predicate="skos:exactMatch", object="mesh:C073260")

    engine = create_engine("sqlite://")
    SQLModel.metadata.create_all(engine)

    # Add edges to the database
    with Session(engine) as session:
        session.add_all([e1, e2])
        session.commit()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(Edge.subject == "CHEBI:135122")
        edges = session.exec(statement).all()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(
            Edge.subject == Reference(prefix="CHEBI", identifier="135125")
        )
        edges = session.exec(statement).all()

If you want to have a field that has an optional reference, you can do it as expected
using the same column type, such as for the ``author`` field below:

.. code-block:: python

    class Edge(SQLModel, table=True):
        id: int | None = Field(default=None, primary_key=True)
        subject: Reference = Field(sa_column=get_reference_sa_column())
        predicate: Reference = Field(sa_column=get_reference_sa_column())
        object: Reference = Field(sa_column=get_reference_sa_column())
        author: Reference | None = Field(sa_column=get_reference_sa_column())

If you want to have a field with a list of references, you can do it using
:func:`get_reference_list_sa_column`:

.. code-block:: python

    class Record(SQLModel, table=True):
        id: int | None = Field(default=None, primary_key=True)
        authors: list[Reference] = Field(
            default_factory=list, sa_column=get_reference_list_sa_column()
        )

This also works how you would expect if you don't use a default_factory or if you choose
to make it optional ``list[Reference] | None``.

Using :mod:`sqlalchemy`
=======================

SQLAlchemy is a combine high- and mid-level database abstraction layer and
object-relational mapping. It has more opportunities for configuration over SQLModel.

The following example re-formulates the above example for SQLModel using the slightly
more direct SQLAlchemy formulation:

.. code-block:: python

    from curies import Reference
    from sqlalchemy import Column, Integer, String, create_engine
    from sqlalchemy.orm import DeclarativeBase, Session


    class Base(DeclarativeBase):
        pass


    class Edge(Base):
        __tablename__ = "edge"

        id = Column(Integer, primary_key=True)

        subject = get_reference_sa_column()
        predicate = get_reference_sa_column()
        object = get_reference_sa_column()


    e1 = Edge(subject="CHEBI:135122", predicate="skos:exactMatch", object="mesh:C073738")
    e2 = Edge(subject="CHEBI:135125", predicate="skos:exactMatch", object="mesh:C073260")

    engine = create_engine("sqlite://")
    Base.metadata.create_all(engine)

    # Add edges to the database
    with Session(engine) as session:
        session.add_all([e1, e2])
        session.commit()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(Edge.subject == "CHEBI:135122")
        edges = session.exec(statement).all()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(
            Edge.subject == Reference(prefix="CHEBI", identifier="135125")
        )
        edges = session.exec(statement).all()

In the following example, SQLAlchemy is used with a composite column, where the prefix
and identifier are given their own column, then a :class:`sqlalchemy.Composite` object
is used to tell the ORM to map both the prefix and identifier together into a single
column, that exposes an appropriate :class:`curies.Reference` class.

.. code-block:: python

    from curies import Reference
    from sqlalchemy import Column, Integer, String, create_engine
    from sqlalchemy.orm import DeclarativeBase, Session


    class Base(DeclarativeBase):
        pass


    class Edge(Base):
        __tablename__ = "edge"

        id = Column(Integer, primary_key=True)

        subject_prefix = Column(String, nullable=False)
        subject_identifier = Column(String, nullable=False)
        predicate_prefix = Column(String, nullable=False)
        predicate_identifier = Column(String, nullable=False)
        object_prefix = Column(String, nullable=False)
        object_identifier = Column(String, nullable=False)

        subject = get_reference_sa_composite(subject_prefix, subject_identifier)
        predicate = get_reference_sa_composite(predicate_prefix, predicate_identifier)
        object = get_reference_sa_composite(object_prefix, object_identifier)


    e1 = Edge(subject="CHEBI:135122", predicate="skos:exactMatch", object="mesh:C073738")
    e2 = Edge(subject="CHEBI:135125", predicate="skos:exactMatch", object="mesh:C073260")

    engine = create_engine("sqlite://")
    Base.metadata.create_all(engine)

    # Add edges to the database
    with Session(engine) as session:
        session.add_all([e1, e2])
        session.commit()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(Edge.subject == "CHEBI:135122")
        edges = session.exec(statement).all()

    # Query for edges with a given subject, by string
    with Session(engine) as session:
        statement = select(Edge).where(
            Edge.subject == Reference(prefix="CHEBI", identifier="135125")
        )
        edges = session.exec(statement).all()

Similarly to above, lists can be constructed using :func:`get_reference_list_sa_column`:

.. code-block:: python

    class Record(Base):
        __tablename__ = "record"

        id = Column(Integer, primary_key=True)

        authors = get_reference_list_sa_column()
"""

from __future__ import annotations

from typing import Any, ClassVar

import sqlalchemy
from sqlalchemy import Column
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.orm import Composite, composite
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy.types import JSON, TEXT, TypeDecorator

from curies import Reference

__all__ = [
    "SAReferenceListTypeDecorator",
    "SAReferenceTypeDecorator",
    "get_reference_list_sa_column",
    "get_reference_sa_column",
    "get_reference_sa_composite",
]


class SAReferenceListTypeDecorator(TypeDecorator[list[Reference]]):
    """A SQLAlchemy type decorator for a list of :mod:`curies.Reference`."""

    impl: ClassVar[type[TypeEngine[str]]] = JSON  # type:ignore[misc]
    #: Set SQLAlchemy caching to true
    cache_ok: ClassVar[bool] = True  # type:ignore[misc]

    def process_bind_param(
        self, value: str | Reference | list[Reference] | None, dialect: Dialect
    ) -> list[str] | None:
        """Convert the Python object into a database value."""
        if value is None:
            return None
        if isinstance(value, str):
            return [value]
        elif isinstance(value, Reference):
            return [value.curie]
        else:
            return [v.curie for v in value]

    def process_result_value(
        self, value: list[str] | None, dialect: Dialect
    ) -> list[Reference] | None:
        """Convert the database value into a Python object."""
        if value is None:
            return None
        return [Reference.from_curie(v) for v in value]


class SAReferenceTypeDecorator(TypeDecorator[Reference]):
    """A SQLAlchemy type decorator for a :mod:`curies.Reference`."""

    impl: ClassVar[type[TypeEngine[str]]] = TEXT  # type:ignore[misc]
    #: Set SQLAlchemy caching to true
    cache_ok: ClassVar[bool] = True  # type:ignore[misc]

    def process_bind_param(self, value: str | Reference | None, dialect: Dialect) -> str | None:
        """Convert the Python object into a database value."""
        if value is None:
            return None
        if isinstance(value, str):
            return value
        return value.curie

    def process_result_value(self, value: str | None, dialect: Dialect) -> Reference | None:
        """Convert the database value into a Python object."""
        if value is None:
            return None
        return Reference.from_curie(value)

    # TODO what about process literal param?


def get_reference_sa_column(*args: Any, **kwargs: Any) -> sqlalchemy.Column[Reference]:
    """Get a SQLAlchemy column with the type decorator for a :mod:`curies.Reference`.

    :param args: positional arguments, passed to :class:`sqlalchemy.Column`
    :param kwargs: keyword arguments, passed to :class:`sqlalchemy.Column`

    :returns: A column object, parametrized with :class:`curies.Reference`

    For example, this can be used to model a semantic triple, which has a subject
    reference, predicate reference, and object reference like in the following:

    .. code-block:: python

        from curies import Reference
        from curies.database import get_reference_sa_column
        from sqlmodel import Field, SQLModel


        class Edge(SQLModel, table=True):
            id: int | None = Field(default=None, primary_key=True)
            subject: Reference = Field(sa_column=get_reference_sa_column())
            predicate: Reference = Field(sa_column=get_reference_sa_column())
            object: Reference = Field(sa_column=get_reference_sa_column())
    """
    return sqlalchemy.Column(SAReferenceTypeDecorator(), *args, **kwargs)


def get_reference_list_sa_column(*args: Any, **kwargs: Any) -> sqlalchemy.Column[list[Reference]]:
    """Get a SQLAlchemy column with the type decorator for a :list of mod:`curies.Reference`.

    :param args: positional arguments, passed to :class:`sqlalchemy.Column`
    :param kwargs: keyword arguments, passed to :class:`sqlalchemy.Column`

    :returns: A column object, parametrized with list of :class:`curies.Reference`

    For example, this can be used to model an author list like in the following:

    .. code-block:: python

        from curies import Reference
        from curies.database import get_reference_list_sa_column
        from sqlmodel import Field, SQLModel


        class Edge(SQLModel, table=True):
            id: int | None = Field(default=None, primary_key=True)
            authors: list[Reference] = Field(
                default_factory=list, sa_column=get_reference_list_sa_column()
            )
    """
    return sqlalchemy.Column(SAReferenceListTypeDecorator(), *args, **kwargs)


class _ReferenceAdapter(Reference):
    """A wrapper for SQLAlchemy for usage in composite()."""

    def __init__(self, prefix: str, identifier: str) -> None:
        """Initialize the SQLAlchemy model."""
        super().__init__(prefix=prefix, identifier=identifier)


def get_reference_sa_composite(
    prefix_column: Column[str], identifier_column: Column[str], **kwargs: Any
) -> Composite[Reference]:
    """Get a composite for a reference.

    :param prefix_column: The column for the reference's prefix
    :param identifier_column: The column for the reference's identifier
    :param kwargs: keyword arguments passed to :func:`sqlalchemy.orm.composite`

    :returns: A Composite object for a reference

    .. code-block:: python

        from curies import Reference
        from curies.database import get_reference_sa_composite
        from sqlalchemy import Column
        from sqlalchemy.orm import DeclarativeBase


        class Base(DeclarativeBase):
            pass


        class Edge(Base):
            __tablename__ = "edge"

            id = Column(Integer, primary_key=True)

            subject_prefix = Column(String, nullable=False)
            subject_identifier = Column(String, nullable=False)
            predicate_prefix = Column(String, nullable=False)
            predicate_identifier = Column(String, nullable=False)
            object_prefix = Column(String, nullable=False)
            object_identifier = Column(String, nullable=False)

            subject = get_reference_sa_composite(subject_prefix, subject_identifier)
            predicate = get_reference_sa_composite(predicate_prefix, predicate_identifier)
            object = get_reference_sa_composite(object_prefix, object_identifier)
    """
    return composite(_ReferenceAdapter, prefix_column, identifier_column, **kwargs)
